#!/bin/env python
import threading
import os
from time import time
import signal
import queue
import uuid
import errno
import cs.Tools.Constants as Constants
from cs.Tools.utils import thread_id_string, adler32, get_bw, dir_is_locked
from cs.Tools.LogConfig import enable_file_logging


class TransferInfo:
    def __init__(self, process, file_meta_data, retry, start_time, local_file_size,
            db_checksum, db_file_size, file_health):
        self.process = process
        self.file_info = file_meta_data
        self.retry = retry
        self.start_time = start_time
        self.local_file_size = local_file_size
        self.db_checksum = db_checksum
        self.db_file_size = db_file_size
        self.file_health = file_health


class CopyThread(threading.Thread):

    def __init__(self, conf, event, dbLock, CopyQueue, DeleteQueue, ClearQueue,
            DdmMonitoringQueue):

        threading.Thread.__init__(self,name="CopyThread")

        self.conf = conf
        self.event = event
        self.dbLock = dbLock
        self.CopyQueue = CopyQueue
        self.DeleteQueue = DeleteQueue
        self.ClearQueue = ClearQueue
        self.DdmMonitoringQueue = DdmMonitoringQueue

        # State variables
        self.simple_transfer_failures = 0
        self.problematic_transfer_failures = 0
        self.files_copied = 0

        self.exitFlag = False
        self.CopyList = []
        self.copy_lock = threading.Lock()
        self.db = None
        self.dbFlag = True
        self.ddmQueueFullSince = None #timestamp indicating since when the queue is full

        self.filenameparser = None

        # get logger to files, (Will also log to ERS)
        self.logger = enable_file_logging("CopyThread","CopyLog.out", conf.LogDir, conf.LogLevel)


    def setFilenameParser(self, parser):
        self.filenameparser = parser

    def run(self):
        self.logger.info(thread_id_string())

        while(1):
            # When time to exit, do not take new files to be copied, but finish ongoing copy processes
            with self.copy_lock:
                if self.exitFlag and not self.CopyList: break

            self.logger.debug('Size of CopyQueue: %d, Size of DeleteQueue: %d, Size of ClearQueue: %d',
                    self.CopyQueue.qsize(), self.DeleteQueue.qsize(), self.ClearQueue.qsize())

            self.logger.debug('Launch %d parallel copy processes', self.conf.MaxConcurrentCopy)
            bwCurrent = get_bw(self.conf.BWDevice)
            bwUpdate = 6

            while self.getCopyListLength() < self.conf.MaxConcurrentCopy and bwCurrent < self.conf.BWLimit:

                if self.conf.BWLimit:
                    if not bwUpdate:
                        bwCurrent = get_bw(self.conf.BWDevice)
                        bwUpdate = 5
                    else:
                        bwUpdate -= 1

                if self.exitFlag: break
                if self.CopyQueue.qsize() == 0: break

                fmd = self.CopyQueue.get(0) # fmd: file meta data
                self.logger.debug('Got file %s from CopyQueue', fmd.file_name)

                # Check if filesystem was locked while the file was in the CopyQueue
                if dir_is_locked(os.path.dirname(fmd.file_name), self.conf.SrcDirs, self.conf.LockFilePattern):
                    # Put the locked files into the ClearQueue
                    # The Manager will take it when the directory will be unlocked
                    self.ClearQueue.put(fmd)
                    # block until room is available in the queue ?????
                    self.logger.debug('Filesystem locked (it was unlocked for the Manager): do not copy')
                    continue

                remote_file = os.path.join(fmd.remote_dir, os.path.basename(fmd.file_name))
                self.logger.debug('Create directory %s in remote system', fmd.remote_dir)
                self.conf.backend.mkdir(fmd.remote_dir, fmd.eos_instance, self.logger)

                # consistency checks
                try:
                    local_filesize = os.stat(fmd.file_name)[6]
                except OSError as exc:
                    if exc.errno == errno.ENOENT:
                        # The file has been removed by an external cause.
                        # This is bad but this is not a reason to crash.
                        self.logger.warning('error getting local file size %s: no such file', fmd.file_name)
                        # delete associated file (ignoring errors)
                        try:
                            os.remove(fmd.file_name + Constants.tobecopied_ext)
                        except:
                            pass
                        self.ClearQueue.put(fmd)
                        continue
                    else:
                        raise exc

                dbchecksum = dbfilesize = filehealth = None

                if self.db:
                    with self.dbLock:
                        dbchecksum, dbfilesize, filehealth = self.db.FileInfo(fmd.file_name)

                if filehealth == 'TRUNCATED':
                    ## Fill the database with local info
                    dbchecksum = adler32(fmd.file_name)
                    dbfilesize = local_filesize
                    with self.dbLock:
                        self.db.UpdateTruncated(fmd.file_name, dbfilesize, dbchecksum)

                if self.conf.ComputeChecksumFromLocalFile:
                    dbchecksum = adler32(fmd.file_name)

                valid = True

                if filehealth:
                    valid &= (dbfilesize == local_filesize)

                if valid:
                    try:
                        copyproc = self.conf.backend.backgroundcopy(
                                fmd.file_name, remote_file,
                                fmd.eos_instance, self.logger)
                    except:
                        self.logger.critical('error starting copy: %s', fmd.file_name)
                        raise

                    try:
                        os.rename(fmd.file_name + Constants.tobecopied_ext,
                                  fmd.file_name + Constants.copying_ext)
                    except OSError as exc:
                        if exc.errno == errno.ENOENT:
                            # The file has been removed by an external cause.
                            # This is bad but this is not a reason to crash.
                            self.logger.warning('error renaming %s: no such file',
                                    fmd.file_name + Constants.tobecopied_ext)
                        else:
                            raise exc

                    with self.copy_lock:
                        self.CopyList.append(TransferInfo(copyproc, fmd, 0, time(),
                                local_filesize, dbchecksum, dbfilesize, filehealth))


                else:
                    self.logger.warning('File %s is invalid for copy', fmd.file_name)
                    os.rename(fmd.file_name + Constants.tobecopied_ext,
                              fmd.file_name + Constants.problematic_ext)
                    self.problematic_transfer_failures += 1
                    self.ClearQueue.put(fmd)

            # Check copy processes status
            self.statusCheck()

            self.event.wait(self.conf.CopyThreadEventTimeout)

        self.logger.info('CopyThread exited')


    # Check copy processes status
    def statusCheck(self):
        with self.copy_lock:
            transfer_list = self.CopyList[:]

        for transfer in transfer_list:
            status = self.conf.backend.copystate(transfer.process[0])
            pid = transfer.process[1]
            datafile = transfer.file_info.file_name
            eosinstance = transfer.file_info.eos_instance
            remote_file = os.path.join(transfer.file_info.remote_dir, os.path.basename(datafile))

            # Status None = copy process not yet finished
            if status == None:
                self.logger.debug('Copy of file %s not yet finished', datafile)

                # Status can not be None for ever: after a while kill and retry #####
                if self.conf.TransferTimeout and (time()-transfer.start_time) > self.conf.TransferTimeout:
                    os.kill(pid, signal.SIGKILL)
                    msg = f'copy process for {datafile} was taking too long: kill it and retry'
                    self.handleUnsuccCopy(transfer, msg, 'TIMEOUT')

                # If the FS got locked, stop the copy
                if self.conf.StopOngoingTransferAfterLock and dir_is_locked(os.path.dirname(datafile),
                        self.conf.SrcDirs, self.conf.LockFilePattern):

                    self.logger.info('File %s is being copied, but FS is now locked. Killing and postponing the copy.', datafile)

                    #kill process
                    os.kill(pid, signal.SIGKILL)

                    #remove stale file from distant storage
                    self.conf.backend.remove(remote_file, eosinstance, self.logger)

                    #rename helper file
                    os.rename(datafile + Constants.copying_ext,
                              datafile + Constants.tobecopied_ext)

                    #remove from copylist
                    with self.copy_lock:
                        self.CopyList.remove(transfer)


                    # Register a simple transfer failure
                    self.simple_transfer_failures += 1

                    #send to the manager
                    self.ClearQueue.put(transfer.file_info)

            # Status 0 = successfull copy
            elif status == 0:
                self.logger.debug('copy process stdout: %s',
                        self.conf.backend.copystdout(transfer.process[0]))

                # Compare size of local and remote file
                local_filesize = transfer.local_file_size

                remote_filesize, checksum = self.conf.backend.sizechecksum(
                        remote_file, eosinstance, self.logger)

                success = (local_filesize == remote_filesize)
                if not success:
                    failure_reason = 'local and remote files have different sizes: {}, {}'.format(local_filesize, remote_filesize)
                    self.handleUnsuccCopy(transfer, failure_reason, 'FILEINTEGRITY')
                    continue

                if remote_filesize:
                    # It is possible that the backend experienced an issue
                    if checksum is None:
                        self.logger.warning('backend sizechecksum error for file %s: '
                                'filesize=%s, checksum=%s ; will retry later',
                                remote_file, str(remote_filesize), str(checksum))
                        # retry later
                        continue

                    ### If DB is available, compare also checksum and
                    ### db filesize
                    if transfer.file_health:
                        success = (remote_filesize == transfer.db_file_size)
                        if not success:
                            failure_reason = 'remote file size differs from filesize in database'
                            self.handleUnsuccCopy(transfer, failure_reason, 'FILEINTEGRITY')
                            continue

                        success = (transfer.db_checksum.lower() == checksum.lower())
                        if not success:
                            failure_reason = 'remote file checksum differs from checksum in database'
                            self.handleUnsuccCopy(transfer, failure_reason, 'FILEINTEGRITY')
                            continue
                    ### If only (locally calculated) checksum is available,
                    ### we skip the DB size check
                    elif transfer.db_checksum:
                        success = (transfer.db_checksum.lower() == checksum.lower())
                        if not success:
                            failure_reason = 'local file checksum differs from checksum in database'
                            self.handleUnsuccCopy(transfer, failure_reason, 'FILEINTEGRITY')
                            continue

                # If we reached this point, the copy has been successfull
                self.logger.info('copy successful: %s', datafile)

                # Copy successfull: update Oracle database, if connection is fine
                # File table: file transferstate from ONDISK to TRANSFERRED
                # Lumiblock table: if the case, lumiblock state from CLOSED to TRANSFERRED
                # Run table: if the case, run state from CLOSED to TRANSFERRED
                if self.db:
                    self.logger.info('Update database')
                    self.dbLock.acquire()
                    self.dbFlag = self.db.Transfer(datafile, remote_file)
                    self.dbLock.release()
                elif self.conf.DBURL:
                    self.logger.critical('No connection to database: database will not be updated for file %s', datafile)

                # Copy successfull: rename .data.COPYING in .data.COPIED
                try:
                    os.rename(datafile + Constants.copying_ext,
                              datafile + Constants.copied_ext)
                except OSError as exc:
                    if exc.errno == errno.ENOENT:
                        self.logger.warning('error renaming %s: no such file', datafile + Constants.copying_ext)
                    else:
                        raise exc

                if self.conf.DdmMonitoringEnabled:
                    self.publishInDdmMonitoring(
                            os.path.basename(datafile),  # filename
                            local_filesize,  # filesize
                            int(transfer.start_time * 1000000),  # timestamp in us for start of copy
                            int(time() * 1000000),  # timestamp in us for end of copy
                            'DONE',  # copy status
                            )

                # Copy successfull
                if self.conf.DeleteEnabled:
                    self.DeleteQueue.put(transfer.file_info)
                else:
                    self.skipDelete(transfer.file_info)

                # Remove the copy process from CopyList
                with self.copy_lock:
                    self.CopyList.remove(transfer)

                self.files_copied += 1

            else:
                warning_msg = 'copy process failure; file: %s; exit code: %s; process output:\n%s' % (
                        datafile, str(status), self.conf.backend.copystdout(transfer.process[0]))
                self.handleUnsuccCopy(transfer, warning_msg, 'PROCESS')


    def handleUnsuccCopy(self, transfer, reason, short_reason):
        """
        Handle unsuccessful copies: status = 1, or status = 0 but local and
        remote filesizes don't match, or status = None for too much time

        transfer -- Transfer object
        reason -- warning message for local logger
        short_reason -- string describing the error, used for publication to DDM
                  monitoring. Keep it short and explicit.
        """
        datafile = transfer.file_info.file_name

        self.logger.warning('copy failure: %s; reason: %s', datafile, reason)

        if self.exitFlag:
            with self.copy_lock:
                self.CopyList.remove(transfer)

            return

        self.logger.info('Handle unsuccessful copy of file %s', datafile)
        nretry = transfer.retry
        eosinstance = transfer.file_info.eos_instance
        local_filesize = transfer.local_file_size
        remote_file = os.path.join(transfer.file_info.remote_dir, os.path.basename(datafile))
        self.logger.debug('File was retried for copying on the spot %d times (%d allowed)',
                nretry, self.conf.MaxCopyImmediateRetry)

        # Unsuccessfull copy: retry up to MaxCopyImmediateRetry times
        if nretry < self.conf.MaxCopyImmediateRetry:
            self.logger.debug('Remove file from remote system')
            self.simple_transfer_failures += 1
            self.conf.backend.remove(remote_file, eosinstance, self.logger)

            if self.exitFlag:
                return

            if self.conf.DdmMonitoringEnabled:
                self.publishInDdmMonitoring(
                        os.path.basename(datafile),  # filename
                        local_filesize,  # filesize
                        int(transfer.start_time * 1000000),  # timestamp in us for start of copy
                        int(time() * 1000000),  # timestamp in us for end of copy
                        short_reason,  # copy status
                        )

            self.logger.info('Retry to copy file on the spot again')
            copyproc = self.conf.backend.backgroundcopy(datafile, remote_file,
                    eosinstance, self.logger)
            with self.copy_lock:
                idx = self.CopyList.index(transfer)
                self.CopyList[idx].process = copyproc
                self.CopyList[idx].retry += 1
                self.CopyList[idx].start_time = time()

        else:
            ##### Mark file as problematic, don't delete it and for now give up to copy #####
            ##### Problematic files will be retaken from the Manager with a delay #####
            self.logger.info('Too many retries on the spot: mark file as problematic %s', datafile)
            try:
                os.rename(datafile + Constants.copying_ext,
                          datafile + Constants.problematic_ext)
                self.problematic_transfer_failures += 1
            except OSError as exc:
                if exc.errno == errno.ENOENT:
                    # The file has been removed by an external cause.
                    # This is bad but this is not a reason to crash.
                    self.logger.warning('error renaming %s: no such file', datafile + Constants.copying_ext)
                else:
                    raise exc
            if self.conf.DdmMonitoringEnabled:
                self.publishInDdmMonitoring(
                        os.path.basename(datafile),  # filename
                        local_filesize,  # filesize
                        int(transfer.start_time * 1000000),  # timestamp in us for start of copy
                        int(time() * 1000000),  # timestamp in us for end of copy
                        'FAILED_{}'.format(short_reason),  # format agreed with DDM
                        )

            self.ClearQueue.put(transfer.file_info)
            with self.copy_lock:
                self.CopyList.remove(transfer)

    def getCopyListLength(self):
        with self.copy_lock:
            return len(self.CopyList)

    def copyExit(self):
        self.exitFlag = 1
        self.logger.info('Exit signal received')

    def setDB(self, db):
        self.db = db
        self.logger.debug('Set connection to database as given by main thread')

    def getDBFlag(self):
        self.logger.debug('Communicate to main thread if connection to database exists')
        flag = self.dbFlag
        self.dbFlag = True
        return flag

    def skipDelete(self, file_entry):
        # Send the file back to the ManagerThread where it will be removed
        # from the CopyFileList. The existence of the .COPIED file prevents
        # it from being retransferred.
        self.ClearQueue.put(file_entry)

    def publishInDdmMonitoring(self, filename, filesize, copy_start_ts_us,
            copy_end_ts_us, status):

        # for filename data21_comm.00390617.calibration_SCTNoise.daq.RAW._lb0000._SFO-7._0006.data
        # scope = data21_comm
        # dataset = data21_comm.00390617.calibration_SCTNoise.daq.RAW
        # for SFOFileNameParser scope is projecttag and dataset is dataset
        scope = ''
        dataset = ''
        if self.filenameparser:
            try:
                parsed_fn = self.filenameparser(filename)
                scope = parsed_fn.ProjectTag()
                dataset = parsed_fn.Dataset()
            except:
                self.logger.critical('error parsing filename: %s', filename)
                raise

        d = {
                'uuid': str(uuid.uuid4()).replace('-', '').lower(),
                'account': '',
                'dataset': dataset,
                'scope': scope,
                'eventType': 'sfo2eos',
                'clientState': status,
                'timeStart': copy_start_ts_us,
                'timeEnd': copy_end_ts_us,
                'filename': filename,
                'filesize': filesize,
                'remoteSite': 'CERN-PROD_TZDISK',
                'localSite': 'CERN-P1_SFO'
        }

        try:
            self.DdmMonitoringQueue.put_nowait(d)
            if self.ddmQueueFullSince is not None:
                self.ddmQueueFullSince = None
                self.logger.warning('DdmMonitoringQueue is no longer full')
        except queue.Full:
            # If the network connection (gateway) or the server cannot cope with
            # the request rate, it's better to discard updates here. Otherwise
            # the memory consumption will increase indefinitely.

            # We had problems recovering after connection loss issue:
            # here the queue remain full forever (18 days). So I changed the
            # logic: we log transition from "not full" to full and from full to
            # "not full". And we log a critical if this situation lasts for more
            # than X minutes
            # The DdmMonitoring logfile will contain more details on the problem

            if self.ddmQueueFullSince is None:
                # first occurrence: log warning
                self.logger.warning('DdmMonitoringQueue is full: discarding request for %s', filename)
                self.ddmQueueFullSince = time()
            else:
                self.logger.debug('DdmMonitoringQueue is full: discarding request for %s', filename)

                if (time() - self.ddmQueueFullSince) > 300:
                    self.logger.critical('DdmMonitoringQueue is full for more than 300 seconds: clearing queue. '
                            'Check DdmMonitoring logs and check connection to remote server (with ddm.test.py)')
                    with self.DdmMonitoringQueue.mutex:
                        self.DdmMonitoringQueue.queue.clear()
                    # At the next self.DdmMonitoringQueue.put attempt,
                    # ddmQueueFullSince will be reset so we don't log critical
                    # forever
