#!/bin/env python
import threading
import os
import errno
import importlib
from time import time
import re
import cs.Tools.Constants as Constants
from cs.Tools.utils import thread_id_string
from cs.Tools.utils import getmountpoint
from cs.Tools.utils import dir_is_locked
from cs.Tools.LogConfig import enable_file_logging
from queue import Empty as QueueEmpty

class DeleteThread(threading.Thread):

    def __init__(self, conf, event, dblock, deletequeue, clearqueue):

        threading.Thread.__init__(self, name="DeleteThread")

        self.conf = conf
        self.event = event
        self.dblock = dblock
        self.deletequeue = deletequeue
        self.clearqueue = clearqueue
        self.files_deleted = 0
        self.migration_waitroom = {}
        self.exitflag = 0
        self.db = None
        self.dbflag = True
        self.deletenow = False
        self.last_fs_usage_check_time = 0
        self.fs_usages = {}

        # get logger to files, (Will also log to ERS)
        self.logger = enable_file_logging("DeleteThread", "DeleteLog.out", conf.LogDir, conf.LogLevel)

        self.migration_module = None
        if self.conf.MigrationModuleName:
            pymod = importlib.import_module(
                    f"cs.Tools.MigrationCheckers.{self.conf.MigrationModuleName}")
            self.migration_module = pymod.MigrationChecker(
                    self.conf.MigrationModuleConf, self.logger)


    def run(self):
        self.logger.info(thread_id_string())

        while True:
            self.event.wait(self.conf.DeleteThreadEventTimeout)
            if self.exitflag: break
            self.loopIteration()

        self.logger.info('DeleteThread exited')


    def loopIteration(self):
        self.logger.debug('queue size: %d', self.deletequeue.qsize())
        try:
            fmd = self.deletequeue.get_nowait()
        except QueueEmpty:
            return

        filename = fmd.file_name
        self.logger.debug('got file from queue: %s', filename)

        # Check exclude regex
        if self.conf.DeleteExcludeFileRegex is not None:
            if re.match(self.conf.DeleteExcludeFileRegex, filename):
                self.logger.debug('not deleting file matching exclude regex: %s', filename)
                self.skipDelete(fmd)
                return

        # Do not delete file if it's too recent
        if self.conf.DeleteMinFileAge != 0:
            if (time() - fmd.modification_time) < self.conf.DeleteMinFileAge:
                self.logger.info('file too young to be deleted yet: %s', filename)
                self.deletequeue.put(fmd)
                return

        directory = os.path.dirname(filename)

        # Check directory locking
        if not self.conf.DeleteIgnoreLock:
            if dir_is_locked(directory, self.conf.SrcDirs, self.conf.LockFilePattern):
                self.logger.debug('file system is locked, do not delete yet: %s', filename)
                self.deletequeue.put(fmd)
                return

        # Check file system usage
        fs_usage = self.getFileSystemUsage(filename)

        if fs_usage >= self.conf.DeleteHighCriticalMark: self.deletenow = True
        if fs_usage <= self.conf.DeleteLowCriticalMark: self.deletenow = False

        if self.deletenow:
            self.logger.info('critical deletion: %s', filename)
            self.delete(fmd)
            return

        lowwatermark = self.getLowWaterMark(directory)
        self.logger.debug('usage/lowwatermark for %s: %f/%f', filename, fs_usage, lowwatermark)

        if fs_usage < lowwatermark:
            self.deletequeue.put(fmd)
            self.logger.debug('filesystem usage is below threshold, do not delete: %s', filename)
            return

        # If migration is not required, this is it
        if not self.conf.MigrationCheck:
            self.delete(fmd)
            return

        # Migration
        if not self.checkMigrationDelayExpired(filename):
            self.logger.debug('do not check for migration yet: %s', filename)
            self.deletequeue.put(fmd)
            return

        self.logger.debug('checking migration: %s', filename)
        remote_file = os.path.join(fmd.remote_dir, os.path.basename(filename))

        if self.checkMigration(remote_file, fmd.eos_instance):
            self.logger.debug('file migrated: %s', filename)
            self.delete(fmd)
        else:
            self.logger.debug('file not yet migrated: %s', filename)
            self.deletequeue.put(fmd)


    def checkMigration(self, remotefile, eosinstance):
        if self.migration_module is None:
            return self.conf.backend.migrated(remotefile, eosinstance, self.logger)
        else:
            return self.migration_module.isMigrated(remotefile)


    def checkMigrationDelayExpired(self, filename):
        if self.conf.MigrationCheckDelay == 0:
            return True
        if filename not in self.migration_waitroom:
            try:
                transfer_completed_time = os.path.getctime(filename + Constants.copied_ext)
            except OSError as exc:
                if exc.errno == errno.ENOENT:
                    self.logger.warning('error accessing ctime for %s: no such file, using now for transfer time',
                            filename + Constants.copied_ext)
                    transfer_completed_time = time()
                else:
                    raise exc
            self.migration_waitroom[filename] = transfer_completed_time
        else:
            transfer_completed_time = self.migration_waitroom.get(filename)
        if (time() - transfer_completed_time) > self.conf.MigrationCheckDelay:
            return True
        return False


    def deleteExit(self):
        if self.migration_module:
            self.migration_module.stop()
        self.exitflag = 1
        self.logger.info('exit signal received')


    def setDB(self,db):
        self.logger.debug('setting database')
        self.db = db


    def getDBFlag(self):
        flag = self.dbflag
        self.logger.debug('getDBFlag: %s', flag)
        self.dbflag = True
        return flag


    def delete(self, fmd):
        '''
        fmd: FileMetaData object
        '''
        self.logger.info('deleting file %s', fmd.file_name)

        try:
            os.remove(fmd.file_name)
        except OSError as exc:
            if exc.errno == errno.ENOENT:
                # The file has been removed by an external cause.
                # This is bad but this is not a reason to crash.
                self.logger.warning('error deleting %s: no such file', fmd.file_name)
            else:
                raise exc

        try:
            os.remove(fmd.file_name + Constants.copied_ext)
        except OSError as exc:
            if exc.errno == errno.ENOENT:
                # The .COPIED file has been removed by an external cause.
                # This is bad but this is not a reason to crash.
                self.logger.warning('error deleting %s: no such file', fmd.file_name + Constants.copied_ext)
            else:
                raise exc

        if self.db:
            self.logger.info('update database: deletion of %s', fmd.file_name)
            with self.dblock:
                self.dbflag = self.db.Deletion(fmd.file_name)
        elif self.conf.DBURL:
            self.logger.warning('no connection to database, database will not be updated for deletion of %s', fmd.file_name)

        #increment files deleted, used by the Information Service Thread
        self.files_deleted += 1

        # Send back to manager thread for final cleaning
        self.clearqueue.put(fmd)

        try:
            del self.migration_waitroom[fmd.file_name]
        except KeyError:
            # There are several cases in which the files are not in the
            # migration waiting room: migration check is disabled, high critical
            # mark was reached, etc.
            pass


    def skipDelete(self, file_entry):
        # Send the file back to the ManagerThread where it will be removed
        # from the CopyFileList. The existence of the .COPIED file prevents
        # it from being retransferred.
        self.clearqueue.put(file_entry)


    def getLowWaterMark(self, directory):
        # Get the entry in LowWaterMark associated with the given directory
        idx = None
        for i, srcdirs_item in enumerate(self.conf.SrcDirs):
            if directory == srcdirs_item or directory.startswith(srcdirs_item):
                idx = i
                break
        if idx is None:
            self.logger.critical('cannot find SrcDirs entry for %s, assuming lowwatermark=100', directory)
            lowwatermark = 100.
        else:
            lowwatermark = self.conf.DeleteLowWaterMark[idx]
        return lowwatermark


    def getFileSystemUsage(self, filename):
        diff = time() - self.last_fs_usage_check_time
        if diff >= self.conf.DeleteFileSystemUsagePeriod:
            self.logger.debug('getting file system usage after %d seconds', diff)
            self.fs_usages = self._getFileSystemUsage()
            self.last_fs_usage_check_time = time()
            self.logger.debug('fs usages: %s', str(self.fs_usages))

        try:
            fs_usage = self.fs_usages[getmountpoint(filename)]
        except KeyError:
            self.logger.critical('cannot get fs associated with %s, assuming usage=0', filename)
            fs_usage = 0.
        return fs_usage


    def _getFileSystemUsage(self):
        fs_usages = {}
        for srcdirs_item in self.conf.SrcDirs:
            mountpoint = getmountpoint(srcdirs_item)
            if mountpoint in fs_usages: continue

            try:
                stats = os.statvfs(mountpoint)
            except OSError as e:
                self.logger.error('cannot stat %s: %s', e.filename, e.strerror)
                continue

            fs_usages[mountpoint] = 100. * (1 -
                    float(stats.f_bavail) / float(stats.f_blocks))

        return fs_usages
