#!/usr/bin/env tdaq_python
import threading
import os
from datetime import datetime
import ispy

from cs.Tools.utils import thread_id_string
from cs.Tools.utils import getmountpoint
from cs.Tools.LogConfig import enable_file_logging, make_tdaq_app_name, make_fs_data_identifier, make_is_data_identifier

# --------------------------------
# The XML schema associated with CastorScript's IS classes, defining
# CastorScriptState and FileSystemInfo, is part of the SFOng package, because
# the schema needs to be deployed in to the TDAQ release and CastorScript is not
# a TDAQ project.
# --------------------------------

class InfoServiceThread(threading.Thread):

    def __init__(self, conf, event, copy_thread, delete_thread, manager_thread):
        threading.Thread.__init__(self, name="InfoServiceThread")
        self.event = event
        self.conf = conf
        self.time_created = datetime.utcnow()

        self._copy_thread = copy_thread
        self._delete_thread = delete_thread
        self._manager_thread = manager_thread
        self.exit_flag = False

        # get partition to publish to
        self.ipc_partition = ispy.IPCPartition(conf.ISPartition) #pylint: disable=no-member
        self.partition_connected = True

        # IS Server to publish to
        # Errors comes up with: IS repository, not server
        self.is_server = conf.ISServer

        self.tdaq_app_name = make_tdaq_app_name(conf)
        self.logger = enable_file_logging("InfoServiceThread", "InfoServiceLog.out", conf.LogDir, conf.LogLevel)


    def run(self):
        self.logger.info(thread_id_string())

        if self.ipc_partition.isValid():
            self.logger.info('partition %s is valid: IS publication enabled',
                    self.ipc_partition.name())

        # variable to handle the period between the partition becomes valid
        # and the first publication attempt
        # None means that there is nothing to do
        # A valid time means that publication to IS must be postponed to at least
        # partition_valid_at + wait_after_partition_valid
        partition_valid_at = None

        while not self.exit_flag:
            if not self.partition_connected and self.ipc_partition.isValid():
                self.partition_connected = True
                self.logger.warning('partition %s is now valid: IS publications enabled'
                        ' (in %d seconds)', self.ipc_partition.name(), self.conf.ISWaitAfterParitionValid)
                partition_valid_at = datetime.utcnow()
            elif self.partition_connected and not self.ipc_partition.isValid():
                self.partition_connected = False
                self.logger.warning('partition %s is now invalid: IS publications disabled',
                        self.ipc_partition.name())
                # if partition was "invalidated" during a wait period this must be reset
                partition_valid_at = None

            if self.ipc_partition.isValid():
                # still need to check if we are in the period after partition became valid
                if partition_valid_at is not None:
                    if (datetime.utcnow() - partition_valid_at).total_seconds() > self.conf.ISWaitAfterParitionValid:
                        # we waited long enough
                        partition_valid_at = None

                if partition_valid_at is None:
                    self.send_update()

            self.event.wait(self.conf.ISPublicationPeriod)

        # to reset the uptime just before normal exit
        self.update_cs_info()
        self.logger.info('InfoServiceThread exited')


    def info_service_exit(self):
        self.exit_flag = True
        self.logger.info('Exit signal received')

    def send_update(self):
        self.update_cs_info()
        if self.conf.ISFileSystemInfoEnabled:
            self.update_fs_info()

    def update_fs_info(self):
        fs_data_type = "FileSystemInfo"

        for srcdirs_item in self.conf.SrcDirs:
            mountpoint = getmountpoint(srcdirs_item)
            try:
                fs = os.statvfs(mountpoint)
            except OSError as e:
                self.logger.error('Cannot stat %s: %s', e.filename, e.strerror)
                continue

            fs_data_identifier = make_fs_data_identifier(self.is_server, self.tdaq_app_name, mountpoint)

            try:
                self.logger.debug("creating ISObject: %s, type: %s", fs_data_identifier, fs_data_type)
                is_data = ispy.ISObject(self.ipc_partition, fs_data_identifier, fs_data_type)

                is_data.total_MB = (fs.f_blocks * fs.f_bsize) // 1000000
                is_data.available_MB = (fs.f_bavail * fs.f_bsize) // 1000000
                is_data.occupancy_percent = 100.*(1-float(fs.f_bavail)/float(fs.f_blocks))

                self.logger.debug("checkin of is_data")
                is_data.checkin()
                self.logger.debug("Update sent: %s", str(is_data))
            except Exception as ex:
                self.logger.error("Error occured in update_fs_info(): %s", str(ex))



    def update_cs_info(self):
        is_data_type = "CastorScriptState"
        is_data_identifier = make_is_data_identifier(self.is_server, self.tdaq_app_name)

        try:
            self.logger.debug("creating ISObject: %s, type: %s", is_data_identifier,is_data_type)
            is_data = ispy.ISObject(self.ipc_partition, is_data_identifier, is_data_type)

            is_data.uptime_seconds = self.get_process_uptime_in_seconds()

            #simple
            is_data.files_to_be_copied = self._copy_thread.CopyQueue.qsize()
            is_data.files_to_be_deleted = self._copy_thread.DeleteQueue.qsize()

            #processed values
            is_data.files_copied = self._copy_thread.files_copied
            is_data.files_deleted = 0
            if self._delete_thread:
                is_data.files_deleted = self._delete_thread.files_deleted
            is_data.simple_transfer_failures = self._copy_thread.simple_transfer_failures
            is_data.problematic_transfer_failures = self._copy_thread.problematic_transfer_failures
            is_data.files_being_copied = self._copy_thread.getCopyListLength()

            self.logger.debug("getting the problematic dictionary lock from manager_thread")
            with self._manager_thread.ProblDict_lock:
                is_data.files_currently_problematic = len(self._manager_thread.ProblDict)

            self.logger.debug("Sending state: Uptime: %s, problematic_errors: %s, simple_errors: %s",
                is_data.uptime_seconds, is_data.problematic_transfer_failures, is_data.simple_transfer_failures)

            self.logger.debug("checkin of is_data")
            is_data.checkin()
            self.logger.debug("Update sent: %s", str(is_data))
        except Exception as ex:
            self.logger.error("Error occured in update_cs_info(): %s", str(ex))



    def get_process_uptime_in_seconds(self):
        if self.exit_flag:
            # to reset the uptime just before normal exit
            return 0
        else:
            return (datetime.utcnow() - self.time_created).total_seconds()
