#!/bin/env python
import threading, os
from time import time
import datetime
import pprint
import math
import cs.Tools.Constants as Constants
import errno
import re
from cs.Tools.utils import thread_id_string
from cs.Tools.utils import dir_is_locked
from cs.Tools.LogConfig import enable_file_logging
from functools import cmp_to_key
import pathlib
import cs.Threads.FileMetaData as FileMetaData

class ManagerThread(threading.Thread):

    def __init__(self, conf, event, CopyQueue, DeleteQueue, ClearQueue, dblock):

        threading.Thread.__init__(self, name="ManagerThread")

        self.parser = None
        self.conf = conf
        self.event = event
        self.CopyQueue = CopyQueue
        self.DeleteQueue = DeleteQueue
        self.ClearQueue = ClearQueue

        self.db = None
        self.dblock = dblock
        self.dbflag = True

        self.CopyFileList = [] # all files currently handled by CastorScript
        self.ProblDict = {} # files declared problematic with extra info to manage retries
        self.ProblDict_lock = threading.Lock()
        self.exitFlag = False
        self.Year = ''
        self.Month = ''

        # get logger to files, (Will also log to ERS)
        self.logger = enable_file_logging("ManagerThread", "ManagerLog.out", conf.LogDir, conf.LogLevel)


    def initialize(self):
        # Update current year and month cache
        self.updateDates()

        # Check the validity of all the helper files from previous runs
        self.logger.debug('Checking existing helper files')

        # If deletion is enabled, handle COPIED files first
        if self.conf.DeleteEnabled:
            self.logger.info('Handling existing .COPIED files')
            files_to_delete = []

            for folder in self.conf.SrcDirs:
                pattern = '**/*'+Constants.copied_ext if self.conf.RecursiveSourceFolder else '*'+Constants.copied_ext
                for oldfile in pathlib.Path(folder).glob(pattern):
                    if oldfile.is_dir(): continue
                    datafile = os.path.splitext(oldfile)[0]
                    fmd = self.getFileMetaData(datafile, folder)
                    if fmd:
                        files_to_delete.append(fmd)
                    else:
                        # file was deleted, nothing we can do
                        ManagerThread.cleanHelperFiles(datafile)

            # sort list to delete oldest files first
            files_to_delete.sort(key=cmp_to_key(FileMetaData.fileOrder))
            for fmd in files_to_delete:
                self.logger.debug('Adding file %s to DeleteQueue', fmd.file_name)
                self.CopyFileList.append(fmd.file_name)
                self.DeleteQueue.put(fmd)

            self.logger.debug('Size of DeleteQueue: %d', self.DeleteQueue.qsize())

        self.logger.info('Getting initial list of files to transfer')
        for fmd in self.getCopyFiles():
            # order is important here: first CopyFileList then send to a queue
            self.CopyFileList.append(fmd.file_name)
            if self.conf.CopyEnabled:
                self.logger.debug('Adding file %s to CopyQueue', fmd.file_name)
                self.CopyQueue.put(fmd)
            elif self.conf.DeleteEnabled:
                # do as if the file was copied
                self.logger.debug('Adding file %s to DeleteQueue', fmd.file_name)
                os.rename(fmd.file_name + Constants.tobecopied_ext,
                          fmd.file_name + Constants.copied_ext)
                self.DeleteQueue.put(fmd)
            # else: no copy and no delete is not possible (enforced in CastorScript.py)

        self.logger.debug('Size of CopyQueue: %d', self.CopyQueue.qsize())


    def setDB(self, db):
        self.db = db
        self.logger.debug('set connection to database as given by main thread')


    def getDBFlag(self):
        self.logger.debug('Communicate to main thread if connection to database exists')
        flag = self.dbflag
        self.dbflag = True
        return flag


    def loop(self):
        while not self.exitFlag:
            self.event.wait(self.conf.ManagerThreadEventTimeout)
            self.logger.info('Clear processed files; ClearQueue size: %d', self.ClearQueue.qsize())

            clearCounter = 0

            while clearCounter < self.conf.ManagerThreadClearedFilesPeriod:

                if self.exitFlag: break
                if self.ClearQueue.qsize() == 0: break
                self.logger.debug('clearCounter = %d; ManagerThreadClearedFilesPeriod = %d',
                        clearCounter, self.conf.ManagerThreadClearedFilesPeriod)

                # Get file to be cleared from ClearQueue
                fmd = self.ClearQueue.get(0)

                # Remove clearfile from CopyFileList
                self.CopyFileList.remove(fmd.file_name)
                if not os.path.isfile(fmd.file_name + Constants.problematic_ext):
                    with self.ProblDict_lock:
                        self.ProblDict.pop(fmd.file_name, None)

                self.logger.debug('File %s removed from CopyFileList', fmd.file_name)

                # Count the deleted files
                clearCounter += 1

            self.logger.debug('ClearQueue size: %d', self.ClearQueue.qsize())

            # After deletion of ManagerThreadClearedFilesPeriod files, update
            # the list of files to be copied and the copy queue
            self.logger.info('Searching new files to handle')
            for fmd in self.getCopyFiles():
                self.CopyFileList.append(fmd.file_name)
                if self.conf.CopyEnabled:
                    self.logger.debug('Adding file %s to CopyQueue', fmd.file_name)
                    self.CopyQueue.put(fmd)
                elif self.conf.DeleteEnabled:
                    self.logger.debug('Adding file %s to DeleteQueue', fmd.file_name)
                    # do as if the file was copied
                    os.rename(fmd.file_name + Constants.tobecopied_ext,
                              fmd.file_name + Constants.copied_ext)
                    self.DeleteQueue.put(fmd)
            # else: no copy and no delete is not possible

            self.logger.debug('Size of CopyQueue: %d', self.CopyQueue.qsize())

            # Update current year and month for directory streaming
            self.updateDates()


    def run(self):
        self.logger.info(thread_id_string())

        try:
            self.initialize()
            self.loop()
        except IOError as e:
            self.logger.critical('I/O error for file %s: "%s"; '
                    'ManagerThread is stopping', e.filename, e.strerror)

        self.logger.info('ManagerThread exited')


    # Get the new files to be copied
    def getCopyFiles(self):
        date_file_list = []

        for folder in self.conf.SrcDirs:
            # Consider only unlocked filesystems
            if dir_is_locked(folder, self.conf.SrcDirs, self.conf.LockFilePattern):
                self.logger.debug('Directory "%s" is locked: skipping new files search for now', folder)
                continue

            files = []
            for pattern in self.conf.DataFilePattern:
                if self.conf.RecursiveSourceFolder:
                    pattern = '**/' + pattern
                files += [str(f) for f in pathlib.Path(folder).glob(pattern) if not f.is_dir()]
            self.logger.debug('%d files found in folder "%s" matching "%s"',
                    len(files), folder, pprint.pformat(self.conf.DataFilePattern))

            for filename in files:
                # Protection against bad pattern definition in the previous glob
                if os.path.splitext(filename)[1] in (
                        Constants.tobecopied_ext, Constants.copied_ext,
                        Constants.copying_ext, Constants.problematic_ext):
                    continue

                if filename in self.CopyFileList:
                    self.logger.debug('File %s is already being handled', filename)
                else:
                    self.logger.debug('New file: %s', filename)

                    if self.conf.ExcludeFileRegex is not None and re.match(self.conf.ExcludeFileRegex, filename):
                        self.logger.debug('ignoring file matching exclude regex: %s', filename)
                        continue

                    if os.path.isfile(filename + Constants.copied_ext):
                        # The file does not exist in CopyFileList
                        # but it has an associated .COPIED file:
                        # it happens when deletion is disabled; in this case we
                        # don't want to transfer the file over and over again.
                        self.logger.debug('ignoring file already transferred: %s', filename)
                        continue

                    # Check for minimal file size, if needed
                    if self.conf.MinSizekB:
                        try:
                            filesize = os.path.getsize(filename)
                        except OSError as exc:
                            if exc.errno == errno.ENOENT:
                                # The file has been removed by an external cause.
                                # This is bad but this is not a reason to crash.
                                self.logger.warning('error getting file size for %s: no such file', filename)
                                continue
                            else:
                                raise exc

                        if (filesize / 1024.) < self.conf.MinSizekB:
                            self.logger.debug('File %s is too small: skipping', filename)
                            if self.conf.RemoveSmallFiles:
                                self.logger.debug('File %s is too small: deleting', filename)
                                try:
                                    os.remove(filename)
                                except OSError as ex:
                                    self.logger.warning('Error deleting %s: %s', filename, ex.__repr__())

                            continue

                    # Check if .PROBLEMATIC file is ready to be copied
                    if os.path.isfile(filename + Constants.problematic_ext) \
                            and not self.checkProbl(filename):
                        continue

                    # If there is a connection to a database, check first that the file
                    # already exists in DB. This is to avoid the race condition between
                    # the file creator (e.g. SFOng) and the CastorScript instance: the
                    # file may exist in the folder but the request to create the database
                    # entry still in a queue/thread. If it were to happen, the file could
                    # be transferred and the CastorScript could try to update its status
                    # while it still hasn't been created in the database.
                    if self.conf.DBURL:
                        if self.db:
                            with self.dblock:
                                status, exists = self.db.FileExists(filename)
                                if not status:
                                    self.logger.critical('Error accessing the database')
                                    self.dbflag = False
                                    continue
                                if not exists:
                                    self.logger.warning('Skipping file because no entry in database yet: %s', filename)
                                    continue
                        else:
                            # DB is configured but not working at the moment
                            self.logger.warning('Skipping file because db not working at the moment: %s', filename)
                            continue

                    fmd = self.getFileMetaData(filename, folder)
                    if fmd:
                        date_file_list.append(fmd)
                    else:
                        # file was deleted, nothing we can do
                        ManagerThread.cleanHelperFiles(filename)

        self.logger.debug('Sort new files, with priority to express streams, then to oldest ones')
        date_file_list.sort(key=cmp_to_key(FileMetaData.fileOrder))
        return date_file_list

    @staticmethod
    def cleanHelperFiles(filename):
        try:
            os.remove(filename + Constants.tobecopied_ext)
            os.remove(filename + Constants.copying_ext)
            os.remove(filename + Constants.copied_ext)
            os.remove(filename + Constants.problematic_ext)
        except:
            pass


    # ProblDict is a dictionary of {.PROBLEMATIC file : [timestamp, nretry]}
    def checkProbl(self,ProblFile):
        self.logger.debug('Problematic file: %s; check if it is time to retry the copying', ProblFile)
        with self.ProblDict_lock:
            if ProblFile not in self.ProblDict:
                self.ProblDict[ProblFile] = [time(), 0]
                self.logger.debug('First time: retry')
                return True

            timestamp = self.ProblDict.get(ProblFile)[0]
            nretry = self.ProblDict.get(ProblFile)[1]

        # Try to copy the file again only if the timeout is expired
        if (time()-timestamp) > (self.conf.ProblDelay * math.exp(self.conf.ProblScalingFactor*nretry)):
            nretry += 1
            with self.ProblDict_lock:
                self.ProblDict[ProblFile] = [time(), nretry]
            self.logger.debug('Time to retry (%d times up to now)', nretry)

            # Inform the user if we cannot copy the file and the retry time
            # limit has been reached
            if self.conf.ProblDelay*math.exp(self.conf.ProblScalingFactor*nretry) > self.conf.ProblDelayLimit:
                filesize = os.path.getsize(ProblFile)
                self.logger.critical('Problematic file %s reached the delay retry limit of %d seconds. '
                        'Please have a look! (filesize= %d, transfer_timeout= %d)',
                        ProblFile, self.conf.ProblDelayLimit, filesize, self.conf.TransferTimeout)
            return True
        else:
            self.logger.debug('Last retry was too recent: retry later')
            return False


    def getFileMetaData(self, filename, confdir):
        # confdir is the item in configuration SrcDirs that contains filename
        # filename can be in a sub-directory of confdir

        try:
            mtime = FileMetaData.getMTime(filename)
        except OSError as exc:
            if exc.errno == errno.ENOENT:
                self.logger.warning('error getting mtime for %s: file does not exist')
                return None
            else:
                raise exc

        infoFile = ''
        if os.path.isfile(filename + Constants.tobecopied_ext): infoFile = filename + Constants.tobecopied_ext
        elif os.path.isfile(filename + Constants.copying_ext): infoFile = filename + Constants.copying_ext
        elif os.path.isfile(filename + Constants.problematic_ext): infoFile = filename + Constants.problematic_ext
        elif os.path.isfile(filename + Constants.copied_ext): infoFile = filename + Constants.copied_ext

        if infoFile:
            self.logger.debug('File: %s, reading info from existing helper file %s', filename, infoFile)
            with open(infoFile,'r') as info_file:
                all_lines = info_file.readlines()

            remotedir = ''
            eosinstance = ''

            for line in all_lines:
                if 'CastorDir' in line or 'RemoteDir' in line:
                    # 'CastorDir' is for backward compatibility
                    try:
                        remotedir = line.split()[2]
                    except IndexError:
                        self.logger.critical('cannot parse CastorDir/RemoteDir line in file: %s', infoFile)
                elif 'StageHost' in line or 'EOSInstance' in line:
                    # 'StageHost' is for backward compatibility
                    eosinstance = line.split()[2]

            self.logger.debug('FileMetaData: remote dir = %s, eos instance = %s',
                    remotedir, eosinstance)

            if not infoFile.endswith(Constants.copied_ext):
                os.rename(infoFile, filename + Constants.tobecopied_ext)

            return FileMetaData.FileMetaData(filename, remotedir, eosinstance, mtime)

        try:
            parsed = self.parser(filename)
        except:
            # log to be able to investigate
            self.logger.critical('error parsing filename: %s', filename)
            # backward compatibility: re-raise
            raise

        subdir = ''
        if self.conf.ReplicateTreeStructure and os.path.dirname(filename) != os.path.normpath(confdir):
            subdir = os.path.relpath(os.path.dirname(filename), os.path.normpath(confdir))

        #Prepare dictionary for directory definition
        dirdict = {'year':self.Year,
                   'month':self.Month,
                   'streamtype':parsed.StreamType(),
                   'streamname':parsed.StreamName(),
                   'runnumber':parsed.RunNr(),
                   'filenr':parsed.FileNr(),
                   'lumiblocknr':parsed.LBNr(),
                   'projecttag':parsed.ProjectTag(),
                   'dataset':parsed.Dataset(),
                   'applicationid':parsed.AppId(),
                   'directory':parsed.Directory(),
                   'userdef1':parsed.UserDef1(),
                   'userdef2':parsed.UserDef2(),
                   'userdef3':parsed.UserDef3()}

        self.logger.debug('New file: %s, reading info from configuration', filename)

        # Stream driven case
        def match_pool(drivenPool, parsed):
            return (drivenPool.projecttag.match(parsed.ProjectTag()) and
                    drivenPool.streamtype.match(parsed.StreamType()) and
                    drivenPool.streamname.match(parsed.StreamName()) and
                    drivenPool.lumiblock.match(parsed.LBNr()) and
                    drivenPool.application.match(parsed.AppId()) and
                    drivenPool.directory.match(parsed.Directory()))

        if self.conf.DrivenPool:
            for e in self.conf.DrivenPool:
                if match_pool(e, parsed):
                    self.logger.debug('Stream driven case')
                    remotedir = os.path.normpath(e.targetdir%dirdict)
                    if subdir:
                        remotedir = os.path.join(remotedir, subdir)
                    eosinstance = e.eosinstance
                    self.logger.debug('FileMetaData: remote dir = %s, eos instance = %s',
                            remotedir, eosinstance)
                    fmd = FileMetaData.FileMetaData(filename, remotedir, eosinstance,
                            mtime)
                    fmd.writeToFile(Constants.tobecopied_ext, self.logger)
                    return fmd

        # Normal case
        self.logger.debug('Normal case')
        eosinstance = self.conf.EOSInstance
        remotedir = os.path.normpath(self.conf.CopyDir%dirdict)
        if subdir:
            remotedir = os.path.join(remotedir, subdir)
        self.logger.debug('FileMetaData: remote dir = %s, eos instance = %s',
                remotedir, eosinstance)
        fmd = FileMetaData.FileMetaData(filename, remotedir, eosinstance, mtime)
        fmd.writeToFile(Constants.tobecopied_ext, self.logger)
        return fmd


    # Exit handler: exit in clean way
    def managerExit(self):
        self.exitFlag = 1
        self.logger.info('Exit signal received')


    def updateDates(self):
        self.Year = datetime.date.today().year
        self.Month = datetime.date.today().month


    def setFileParser(self, parser):
        self.parser = parser
